from .LinearRegression.LinearRegression import LinearRegression
from .LinearRegression.MultiVariateLinearRegression import \
    MultivariateLinearRegression
from .PolynomialRegression.PolynomialRegression import PolynomialRegression
